;;**** add profiling of methods

;;;;
;;;; Profile utility from Norvig
;;;;

(defvar *profiled-functions* nil)

(defmacro profile (&rest fn-names)
  "Profile fn-names. With no args, list profiled functions."
  `(mapcar #'profile1
	   (setf *profiled-functions*
		 (union *profiled-functions* ',fn-names))))

(defmacro unprofile (&rest fn-names)
  "Stop profiling fn-names. With no args, stop all profiling."
  `(progn
     (mapcar #'unprofile1 ,(if fn-names `',fn-names '*profiled-functions*))
     (setf *profiled-functions* 
	   ,(if (null fn-names)
		nil
	        `(set-difference *profiled-functions* ',fn-names)))))

(defun profile1 (fn-name)
  "Make the function count how often it is called"
  ;; First save away the old, unprofiled function.
  ;; Then make the name be a new function that increments
  ;; a counter and then calls the original function
  (let ((fn (symbol-function fn-name)))
    (unless (eq fn (get fn-name 'profiled-fn))
      (let ((new-fn (profiled-fn fn-name fn)))
	(setf (symbol-function fn-name) new-fn
	      (get fn-name 'profiled-fn) new-fn
	      (get fn-name 'unprofiled-fn) fn
	      (get fn-name 'profile-time) 0
	      (get fn-name 'profile-count) 0))))
  fn-name)

(defun unprofile1 (fn-name)
  "Make the function stop counting how often it is called"
  (setf (get fn-name 'profile-time) 0)
  (setf (get fn-name 'profile-count) 0)
  (when (eq (symbol-function fn-name) (get fn-name 'profiled-fn))
    ;; normal case: restore unprofiled version
    (setf (symbol-function fn-name) (get fn-name 'unprofiled-fn)))
  fn-name)

;(defun get-fast-time () (get-internal-run-time))
(defmacro get-fast-time () '(get-internal-run-time))

;(defun fast-time-difference (end start) (- end start))
(defmacro fast-time-difference (end start) `(- ,end ,start))

(defun fast-time->seconds (time)
  (float (/ time internal-time-units-per-second)))

(defvar *profile-call-stack* nil)

(defun profile-enter (fn-name)
  (incf (get fn-name 'profile-count))
  (unless (null *profile-call-stack*)
    ;; Time charged against the calling function
    (inc-profile-time (first *profile-call-stack*)
		      (car (first *profile-call-stack*))))
  ;; Put a new entry on the stack
  (push (cons fn-name (get-fast-time)) *profile-call-stack*))

(defun profile-exit (fn-name)
  ;; Time charged against the current function
  (inc-profile-time (pop *profile-call-stack*) fn-name)
  ;; Change the top entry to reflect current time
  (unless (null *profile-call-stack*)
    (setf (cdr (first *profile-call-stack*)) (get-fast-time))))

(defun inc-profile-time (entry fn-name)
  (incf (get fn-name 'profile-time)
	(fast-time-difference (get-fast-time) (cdr entry))))

(defun profiled-fn (fn-name fn)
  "Return a function that increments the count, and times."
  #'(lambda (&rest args)
      ;(profile-enter fn-name)
      (incf (get fn-name 'profile-count))
      (unless (null *profile-call-stack*)
	      ;; Time charged against the calling function
	      (inc-profile-time (first *profile-call-stack*)
				(car (first *profile-call-stack*))))
      ;; Put a new entry on the stack
      (push (cons fn-name (get-fast-time)) *profile-call-stack*)
      (multiple-value-prog1
	(apply fn args)
	;(profile-exit fn-name)
	;; Time charged against the current function
	(inc-profile-time (pop *profile-call-stack*) fn-name)
	;; Change the top entry to reflect current time
	(unless (null *profile-call-stack*)
		(setf (cdr (first *profile-call-stack*)) (get-fast-time))))))

(defun profile-report (&optional (fn-names *profiled-functions*)
				 (key #'profile-time))
  "Report profiling statistics on given functions."
  (let ((total-time (reduce #'+ (mapcar #'profile-time fn-names))))
    (unless (null key)
	    ;;**** I think the copy-list is needed so unprofile isn't messed up
	    (setf fn-names (sort (copy-list fn-names) #'> :key key)))
    (format t "~%Total elapsed time: ~d seconds.~%"
	    (fast-time->seconds total-time))
    (when (< 0 total-time)
      (format t "  Count   Secs Time% Name~%")
      (dolist (name fn-names)
        (format t "~7d ~6,2f  ~3d% ~a~%"
		(profile-count name)
		(fast-time->seconds (profile-time name))
		(round (/ (profile-time name) total-time .01))
		name)))))

(defun profile-count (fn-name) (get fn-name 'profile-count))
(defun profile-time (fn-name) (get fn-name 'profile-time))

(defmacro with-profiling (fn-names &rest body)
  `(progn
     (unprofile ,@fn-names)
     (profile ,@fn-names)
     (setf *profile-call-stack* nil)
     (unwind-protect
	 (progn ,@body)
       (profile-report ',fn-names)
       (unprofile ,@fn-names))))
